# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

######################### We start with some black magic to print on failure.

# Change 1..1 below to 1..last_test_to_print .
# (It may become useful if the test is moved to ./t subdirectory.)

BEGIN { $| = 1; print "1..9\n"; }
END {print "not ok 1\n" unless $loaded;}

use CFITSIO::Simple;
use PDL;
use Data::Dumper;

$loaded = 1;
print "ok 1\n\n";

$test_in  = 'test_in.fits';
$test_in2  = 'test_in2.fits[1]';
$test_in3  = 'test_in3.fits[events]';
$test = 2;

######################### End of black magic.

########################################################################
# Read test data file with multi-dimensional columns
########################################################################

print "$test: Read $test_in2\n";
%data = fits_read_bintbl($test_in2); 
@dims = dims $data{img_raw};
$ok = ($dims[0] == 6);

print $ok ? "ok" : "not ok", " ", $test++, "\n\n";


########################################################################
# Read the header for the first extension of 'test_in.fits'
########################################################################

print "$test:  Read header of $test_in\n";
$hdr = fits_read_hdr($test_in, 2);
$ok = ($hdr->{OBJECT} eq 'TRAPEZIUM');
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Read the list of columns from the first extension of 'test_in.fits'
########################################################################

print "$test: Read $test_in\n";
%data = fits_read_bintbl($test_in, 'p_lsi', 'slot', 'id_string'); 

$ok = (sprintf("%.4f",$data{p_lsi}->at(2,0)) eq '86.3082'
       and $data{slot}->at(1) == 1
       and $data{id_string}->[2] eq 'ACIS-S-6');
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Read the columns from the named extension
########################################################################

print "$test: Read $test_in with extname=FIDPROPS\n";
%data = fits_read_bintbl($test_in, 'p_lsi', 'slot', 'mag_i_avg', 'id_string',
			    {extname => 'FIDPROPS'}); 
$ok = (sprintf("%.4f",$data{p_lsi}->at(2,0)) eq '86.3082'
       and $data{slot}->at(1) == 1
       and $data{id_string}->[2] eq 'ACIS-S-6');
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Minimal statement to read the first bin table in file
########################################################################

print "$test: Read everything in extension 1 of $test_in\n";
%data = fits_read_bintbl($test_in);
$ok = (sprintf("%.4f",$data{p_lsi}->at(2,0)) eq '86.3082'
       and $data{slot}->at(1) == 1
       and $data{id_string}->[2] eq 'ACIS-S-6');
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Now write the data back to a file.  By default, if the file exists then
# it will be over-written
########################################################################

$hdr = fits_read_hdr($test_in2, 2, {list => 1});  # Read header keywords in list format
%data = fits_read_bintbl($test_in2);
print "$test: Write first extension on test_out1.fits\n";
$ok = fits_write_bintbl("test_out1.fits", %data, {keywords => $hdr});
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Add another extension, this time explicitly specifying the
# size of the string columns
########################################################################

print "$test: Append second extension on test_out1.fits\n";
$ok = fits_write_bintbl("test_out1.fits", %data, 
			   { clobber => 1, # This is the DEFAULT
			     append  => 1,
			     tform => { id_string => '9A', # 9 character string
					id_status => '4 string', # 4 char. string
				    } }
			   );
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# And now a complicated example, with mixed Perl lists and Piddles
########################################################################

%data = fits_read_bintbl($test_in);
%data2 = ();
$data2{fltarr} 	  = [1.1,2.2,3.2,4.4];
$data2{intarr} 	  = [4,2,1,8];
$data2{strarr} 	  = ['hi', 'joe', 'jim'];
$data2{slot}   	  = $data{slot};
$data2{mag_i_avg} = $data{mag_i_avg};
$data2{p_lsi}     = $data{p_lsi};	# 2-d piddle

print "$test: Write first extension of test_out2.fits without closing\n";
($ok, $fptr) = 
    fits_write_bintbl("test_out2.fits", %data2,
			{ extname => 'BINTABLE2',
			  close   => 0,

			  # Set some header keywords.  The order is [name, val, type, comment].
			  # Comment and type are optional.  If no type is supplied, 'string' is used.
			  keywords => [['testkey1', 1.2345e2, 'double', 'Test double keyword'],
				       ['testkey2', 1.2345e2, 'float'], 
				       ['testkey3', 1.2345e2],  
				       ['testkey4', 12, 'long', 'Test keyword long']],

			  # Explicitly specify some of the data types.  This is not required,
			  # but useful for efficiency or if you want something besides LONG,
			  # DOUBLE, or STRING.
			  tform => { intarr => 'ushort',
				     strarr => '10 A',
				     fltarr => 'float'},
			  
			  # Set the units
			  tunit => { fltarr => 'gm/cm^2',
				     intarr => 'meters' },
			  
		      }			  
			);
print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Write additional rows of data into the first extension of test_out2.fits
########################################################################

print "$test: Write additional data into first extension of test_out2.fits\n";
$ok = fits_write_bintbl($fptr, %data2, {close => 0});
delete $data2{p_lsi};
$ok &&= fits_write_bintbl($fptr, %data2, {close => 0});
delete $data2{strarr};
$ok &&= fits_write_bintbl($fptr, %data2, {close => 0});
$ok &&= fits_write_bintbl($fptr, {close => 1});

print $ok ? "ok" : "not ok", " ", $test++, "\n\n";


########################################################################
# Copy the first HDU of a FITS file (the PRIMARY) and then copy the
# second.
########################################################################

$hdr = fits_read_hdr($test_in, 2, {list => 1});  # Read header keywords in list format
print "$test: Copy header keywords from a template file \n";
unlink "test_out3.fits";
$ok = fits_write_bintbl("test_out3.fits", %data, {keywords => $hdr,
						  primary_hdu => 'test_in.fits'});
$hdr = fits_read_hdr('test_out3.fits', 1);
$ok &&= ($hdr->{'DATE-OBS'} eq '1999-11-24T05:50:22');
$hdr = fits_read_hdr('test_out3.fits', 2);
$ok &&= ($hdr->{'DATE-END'} eq '1999-11-24T06:23:51');

print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Read an ASCII table
########################################################################

print "$test: Read ASCII table \n";
%data = fits_read_bintbl("test_ascii.fits");
$ok &&= $data{raj2000}->[7] eq '10 10 32.27';

print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

########################################################################
# Read test CXC event file and write back
########################################################################

print "$test: Read $test_in3\n";
($colkeys, %data) = fits_read_bintbl($test_in3, {colkeys=>1}); 
$hdr = fits_read_hdr($test_in3, {list=>1});
fits_write_bintbl('test_out4.fits', %data, {colkeys=>$colkeys, keywords=>$hdr});
$ok = 1;

print $ok ? "ok" : "not ok", " ", $test++, "\n\n";

