package Ska::IO;

use Text::ParseWords;
use Carp;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

require Exporter;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT_OK = qw(
		fits_read_keys
		cfitsio_read_header
		read_param_file
		read_ascii
	       );
%EXPORT_TAGS = (all => \@EXPORT_OK);

$VERSION = '0.01';

# Preloaded methods go here.

our $cfitsio_init;


##*******************************************************************
sub cfitsio_read_header {
##*******************************************************************
# Modified version of P. Ratzlaff's fits_read_header which allows
# for reading keywords from arbitrary extension

    unless ($cfitsio_init) {
	eval "use Astro::FITS::CFITSIO qw( :longnames )";
	$cfitsio_init = 1;
    }

    croak 'Usage: fits_read_header(file|fitsfilePtr,[extension #])'
	unless @_;

    my ($fitsfile, $status);
    my ($n, $left, %header, $key, $value, $comment);

    # Read the arguments.  HDU is optional
    my $file = shift;
    my $hdu  = shift;

    my $obj_passed = 0; # were we passed a fitsfilePtr?

    $status = 0;
    if (UNIVERSAL::isa($file,'fitsfilePtr')) {
	$fitsfile = $file;
	$obj_passed = 1;
    } else {
	# Open the file.
	fits_open_file($fitsfile, $file, Astro::FITS::CFITSIO::READONLY(), $status);
    }

    # Now we have an open file -- check that status is good before
    # proceeding
    unless ($status) {

	# Go to proper HDU if needed

	if ($hdu) {
	    if ($hdu =~ /^\d+$/) { # only digits, so its an extension number
		fits_movabs_hdu($fitsfile, $hdu, undef, $status);
	    } else {		 # Extension name
		fits_movnam_hdu($fitsfile, Astro::FITS::CFITSIO::ANY_HDU(), $hdu, 0, $status);
	    }
	    fits_report_error(*STDERR, $status);
	    return () if ($status);
	}

	# Get the number of fits keywords in primary header
	$fitsfile->get_hdrspace($n, $left, $status);

	# Loop over the keys
	for my $i (1..$n) {
	    last unless $status == 0;

	    $fitsfile->read_keyn($i, $key, $value, $comment, $status);

	    $value =~ s/^\'//;
	    $value =~ s/\'$//;
	    # Store the key/value in a hash
	    $header{$key} = $value;

	    # Store the comments.
	    if (! exists $header{COMMENTS}{$key}) {
		$header{COMMENTS}{$key} = $comment;
	    }
	    # HISTORY keywords, for instance, can be numerous
	    else {
		if (! ref $header{COMMENTS}{$key}) {
		    $header{COMMENTS}{$key} = [ $header{COMMENTS}{$key} ];
		}
		push @{$header{COMMENTS}{$key}}, $comment;
	    }

	}

	# Close the file if we opened it
	$fitsfile->close_file($status) unless $obj_passed;
    }

    # Report an error - may not always want to write to STDERR...
    fits_report_error(*STDERR, $status);

    return wantarray ? (\%header, $status) : \%header;
}

###################################################################################
sub fits_read_keys {
###################################################################################
    my @keys;
    my $file = shift;
    @keys = `fdump $file STDOUT - - prdata- prhead+ page-`;

    my $noquote = "[^']";
    my %keys = ();
    foreach (@keys) {
	if (/(\S+)\s*= '(${noquote}+)'/ or /(\S+)\s*= ([^\/]+)/) {   
	    my $key = $1;
	    my $keyval = $2;
	    $key =~ s/\s+$//;
	    $keyval =~ s/\s+$//;
	    $key =~ s/^\s+//;
	    $keyval =~ s/^\s+//;
	    $keys{$key} = $keyval;
	}
    }
    
    return %keys;
}

###################################################################################
sub read_param_file {
###################################################################################
    my $file = shift;
    my %param = ();
    $file = "gunzip --stdout $file |" if ($file =~ /(\.gz|\.Z|\.z)$/);
    open (PAR, $file) || die "Couldn't open parameter file '$file'\n";
    while (<PAR>) {
	my @fields = quotewords(",", 0, $_);
	$param{$fields[0]} = $fields[3];
    }
    close PAR;

    return %param;
}

##****************************************************************************
sub read_ascii {
##****************************************************************************
    my $file = shift;
    my %opt = (exclude => '\s*#',
	       # include
	       # delim
	       @_
	      );

    my @db;
    my @cols;
    my @vals;
    my $warned_about_mismatch;
    my $min_cols;
    my $i_data = 0;

    open DATA, $file or carp  "read_ascii:: Couldn't open $file\n";

    # Define some pre-compiled regular expressions if needed
    my $exclude = qr/$opt{exclude}/ if $opt{exclude};
    my $include = qr/$opt{include}/ if $opt{include};

    # Caller explicitly spec'd columns
    @cols = @{$opt{cols}} if $opt{cols}; 

    while (<DATA>) {
	$i_data++;
	
	# Skip line if blank or if not included or excluded
	next unless /\S+/;
	next if ($opt{include} && !/$include/);
	next if ($opt{exclude} && /$exclude/);

	# If @cols isn't yet defined, then the first acceptable line defines them
	unless (@cols) {
	    @cols = $opt{delim} ? split $opt{delim} : split;
	    next;
	}

	# Split the line and determine the mininum # of cols from header vs. data
	@vals = $opt{delim} ? split $opt{delim} : split;
	$min_cols = $#vals > $#cols ? $#cols : $#vals;
	
	# Complain just once if there is a mismatch
	unless ($#cols == $#vals or $warned_about_mismatch) {
	    carp "read_ascii:: Mismatch in $file: Data at line $i_data n_data_cols $#vals != n_header_cols $#cols\n";
	    $warned_about_mismatch = 1;
	}
	      
	# Put the available data columns in a hash
	push @db, { map { $cols[$_] => $vals[$_] } (0..$min_cols) };
    }

    return @db;
}

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Ska::IO - I/O utilities for spacecraft/aspect analysis

=head1 SYNOPSIS

  use Ska::IO;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for Ska::IO was created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut

