package Ska::Process;

use 5.008;
use strict;
use warnings;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

use POSIX qw(uname floor strftime);
use Carp;
use File::Glob;
use File::Basename;
use File::Spec;
use FileHandle;
use Cwd;
use Getopt::Long;
use Text::ParseWords;
use Expect::Simple;
use Mail::Send;

require Exporter;

@ISA = qw(Exporter);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT_OK = qw(
		get_archive_files
		run_tool
		make_local_copy
		message
		get_params
		remote_shell_cmd
		parse_config
		send_mail
	       );

%EXPORT_TAGS = ( all => \@EXPORT_OK );

$VERSION = '0.04';


# Preloaded methods go here.

our $arc5gl;

##****************************************************************************
sub get_archive_files {
##****************************************************************************
    my %arg = (version => ['last'],
	       loud    => 0,
	       archive => 1,
	       timeout => 2000,
	       @_);

    my @files;


    my $CWD = cwd;
    $arg{dir} = $arg{dir} || $CWD;  
    chdir $arg{dir};  #  if ($arg{dir});

    # Allow guest user access

    # by default, set the command flag string to be empty 
    my $guestuserflag = qq{};
    
    # if the guestuser argument has been passed, set the
    # string to be the flag
    if ($arg{guestuser}){
	$guestuserflag = ' -guestuser ';
    }

    # Check if file is already available

    unless ((@files = glob($arg{file_glob})) or !$arg{archive}) {

	# Start arc5gl if necessary
	unless ($arc5gl) {
		
	    print "Starting arc5gl\n" if ($arg{loud});
	    $arc5gl = new Expect::Simple { Cmd => "/proj/sot/ska/bin/arc5gl $guestuserflag -stdin",
					   Prompt => 'ARC5GL> ',
					   DisconnectCmd => 'exit',
					   Verbose => $arg{loud},
					   Debug => 0,
					   Timeout => $arg{timeout}
					   };
	}

	print "Getting $arg{prod} from archive\n" if ($arg{loud});
	$arc5gl->send("loud");
	$arc5gl->send("cd $arg{dir}"); #  if ($arg{dir});

	if ($arg{obsid}) {
	    $arc5gl->send("obsid=$arg{obsid}");
	    $arc5gl->send("tstart=");
	    $arc5gl->send("tstop=");
	} elsif ($arg{tstart}) {
	    $arc5gl->send("obsid=");
	    $arc5gl->send("tstart=$arg{tstart}");
	    $arc5gl->send("tstop=$arg{tstop}");
	} else {
	    warn("get_archive_files:: Need obsid or tstart\n");
	    chdir $CWD;
	    return ();
	}
	
	my $version;
	for $version (@{$arg{version}}) {
	    $arc5gl->send("version=$version");
	    $arc5gl->send("get $arg{prod}");
	    @files = glob($arg{file_glob});
	    last if (@files);
	}
	unless (@files) {
	    warn "Could not find $arg{prod} archive file matching $arg{file_glob}\n";
	    chdir $CWD;
	    return ();
	}
    }
	
    if ($arg{gunzip}) {
	foreach (@files) {
	    if (/\.gz$/) {
		system("gunzip -f $_");
		s/\.gz$//;
	    }
	}
    }

#   print "Got $arg{prod} file(s) @files\n" if (@files and $arg{loud});
    chdir $CWD;
    return @files;
}

##***************************************************************************
sub send_mail {
##***************************************************************************
    my %par = @_;
    return unless my $addr_list = $par{addr_list};
    
    my @addr_list = ref($addr_list) eq "ARRAY" ? @{$addr_list} : $addr_list;
    unless ($par{dryrun}) {
	my $msg = Mail::Send->new;
	$msg->subject($par{subject}) if $par{subject};
	$msg->to(@addr_list);
	my $fh = $msg->open;
	print $fh $par{message} if $par{message};
	$fh->close;
    }

    my $diagnostic = "Sent mail with subject '$par{subject}' to @addr_list\n";
    foreach (split "\n", $par{message}) {
	$diagnostic .= "  $_\n";
    }
    print $diagnostic if ($par{loud});

    return $diagnostic;
}
 
############################################################################3
sub run_tool {
############################################################################3
    my $cmd = shift;
    my $cmd_pid;

    # First pull any hash ref off the end.  This specifies parameters for
    # the run_tool subroutine (e.g. 'paste' and 'timeout')
    my %par = (loud => 1);
    if (ref $_[-1] eq 'HASH') {
	%par = %{pop @_};
    }
    local %ENV = CXC::Envs::env($par{ENV}) if $par{ENV};
    
    my %toolpar = @_;		# Get the parameters for the tool to be run

    my ($cmd_root) = split ' ', $cmd;
    my $sep = $par{paste} ? " " : "\n    ";
    system("punlearn $cmd_root") if (keys %toolpar);

    message("Running $cmd$sep") if $par{loud};

    foreach (keys %toolpar) {
	my $arg = $par{unix_params} ? " -$_ $toolpar{$_}" : " $_=$toolpar{$_}";
	$cmd .= $arg;
	message("$arg$sep") if $par{loud};
    }
    message("\n") if $par{loud};

    # Run within eval block to be able to set a local alarm to time out if
    # command doesn't finish in in $par{timeout} seconds
    # See perldoc -f alarm
    eval {
	local $SIG{ALRM} = sub { message("ERROR - $cmd_root command timed out ".localtime()."\n")
				   if $par{loud};
				 die "alarm\n";
			       }; # NB: \n required
	alarm $par{timeout} if $par{timeout};
	
	$cmd_pid = open CMD, "$cmd 2>&1 |"
	  or die "run_tool: ERROR - Could not start $cmd_root command\n";
	while (<CMD>) {
	    message($_) if $par{loud};
	    $ {$par{out}} .= $_ if $par{out}; 
	}
	close CMD;

	alarm 0;
    };

    if ($@) {
	die unless $@ eq "alarm\n"; # propagate unexpected errors
	# timed out.  Find all child processes and kill them one at a time

	message("Caught the timeout alarm at " . localtime() . "\nKilling processes...\n")
	  if $par{loud};

	my $ps_cmd = '/bin/ps -a -o "pid pgid args"';
	my $xfit_gid = getpgrp;
	my @ps = `$ps_cmd`;
	foreach (@ps) {
	    my ($pid, $gid, $cmd) = split;
	    if ($gid =~ /^\d+$/ and $gid == $xfit_gid and $pid != $$) {
		kill 9 => $pid;
		message("Killed pid gid = $_") if $par{loud};
		sleep 2;
	    }
	}
	sleep 5;
	close CMD;

	return;
    }

    1;
}

# Make a local block for message subroutine
BEGIN {
    # Private persistent parameters for message
    my %msg = ( time         => 0, # precede each message with the time
		time_format  => "<<%Y-%b-%d %H:%M>> ",  # use this format (strftime)
		stdout    => 1,	# print to STDOUT
		log_file  => '',
		append    => 0,
		);
    my %log_file_handle;

    ##****************************************************************************
    sub message {
    ##****************************************************************************
	# Optionally print something and log it to a file handle
	my $text = shift;
	my %par = (%msg, @_);  # Set parameters for this instance
	my $file;
	local $_;

	$file = File::Spec->rel2abs($par{log_file}) if $par{log_file};

	if ($par{init}) {
	    delete $par{init};
	    %msg = %par;

	    my $append_mode = $par{append} ? ">>" : ">";

	    # If a log file specified and not already opened, then open it
	    if ($file and not $log_file_handle{$file}) {
		croak "ERROR - Could not open log file $par{log_file}\n"
		  unless ($log_file_handle{$file} = new FileHandle "$append_mode $par{log_file}");
		$log_file_handle{$file}->autoflush(1);
	    }
	    return;
	}

	if ($par{close}) {
	    delete $log_file_handle{$file};
	    return;
	}
	
	my $time = $par{time} ? strftime($par{time_format}, localtime) : '';

	print "$time$text" if $par{stdout};
	foreach (values %log_file_handle) {
	    print { $_ } "$time$text";
	}
    }
}

##****************************************************************************
sub make_local_copy {
#
#  Usage:
#    make_local_copy(in_file, [out_file], [{copy=>1, absolute=>1}])
#    
#  Make a local copy of or link to the specified in_file.  Gunzip
#  it if necessary.  If out_file is not supplied, then use the
#  basename of the in_file.  Don't do anything if the output file is 
#  already there.  Finally, if the 'copy' option is specified in the
#  optional hash reference (last argument), then make an actual copy
#  instead of just linking (if the in_file is not gzipped).
#
##****************************************************************************
    return unless @_;		# Need at least 1 arg
    my $cwd = cwd;

    # First pull any hash ref off the end 
    my %par = (freshen => 1);   # Freshen an outdated local copy
    if (ref $_[-1] eq 'HASH') {
	%par = %{pop @_};
    }

    my $in_file = shift or return;
    return unless (-e $in_file);
    
    my $out_file;
    if (@_) {
	# Explicit file name spec'd
	$out_file = shift;
    } else {
	# Use input file name, and assume it is going in the current directory
	$out_file = basename($in_file);
	$out_file =~ s/\.gz$//;	# Remove any trailing .gz
    }

    $par{absolute} ||= $in_file =~ /^\//;

    # Define absolute file name for $in_file and $out_file
    $in_file = File::Spec->rel2abs($in_file, $cwd);

    unless ($par{absolute}) {
	# Go to the directory spec'd for $out_file
	chdir dirname($out_file) 
	  or croak "ERROR - could not change dir to make local copy of $out_file\n";
	$out_file = basename $out_file;
	# Now make in_file relative to the out_file directory
	$in_file = File::Spec->abs2rel($in_file);
    } else {
	$out_file = File::Spec->rel2abs($out_file, $cwd);
    }

    # Delete outfile if it exists and is older than input file, and 'freshen' par allows
    unlink $out_file if (-e $out_file and ($par{force}
					   or ($par{freshen}
					       and [stat $in_file]->[9] > [stat $out_file]->[9])));

    # Now gunzip or link, unless out_file exists and is older than the input file
    unless (-e $out_file) {
	my $copy_cmd = exists $par{copy} ? "cp" : "ln -s";
	system( $in_file =~ /\.gz$/ ?
		"gunzip --stdout $in_file > $out_file" : 
		"$copy_cmd $in_file $out_file");
	message("Making local file " . File::Spec->rel2abs($out_file) . "\n");
    }

    chdir $cwd;
    return $out_file;
}

##****************************************************************************
sub get_params {
##****************************************************************************
    my $param_file = shift;
    my @uparm = split ':', ($ENV{UPARM} || ".");
    my %par;
    my %dt;
    my $success = 0;
    my $uparm;

    foreach $uparm (@uparm) {
	open (PARAMS, "$uparm/$param_file") || next;
	$success = 1;
	while (<PARAMS>) {
	    next if (m"^\s*#" || /^\s*$/);   
	    chomp;
	    my ($name,$datatype,undef,$val) = quotewords(",", 0, $_);
	    if ($datatype eq "b") {
		$val =~ s/yes/1/i;
		$val =~ s/no/0/i;
	    }
	    $par{$name} = $val if ($val ne 'INDEF' and $val =~ /\S/);
	    ($dt{$name} = $datatype) =~ tr/bris/bfis/;
	}
	close PARAMS;
	last;
    }

    unless ($success) {
	carp "Get_params:: Couldn't find parameter file $param_file\n";
	return;
    }

# Based on model params from param file, construct a list of legal command
# line options, and then GetOptions them.

    my @pfile_args = map { $dt{$_} eq "b" ? "$_!" : "$_=$dt{$_}" } keys %dt;

    GetOptions(\%par, 
	       @_,		# Explicitly commanded to subroutine
	       @pfile_args,	# Derived from parameter file
	       ) ||
		   exit( 1 );
    return (%par);
 }

##****************************************************************************
sub remote_shell_cmd {
##****************************************************************************
    my $rsh = '';
    my ($sysname) = POSIX::uname;
    if ($sysname =~ /linux/i) {
	my %machines = (aldcroft => 'garth');
	$rsh = "rsh " . ($machines{$ENV{USER}} || 'garth');
    }
    return $rsh;
}

##****************************************************************************
sub parse_config {
# 
# Parse the configuration file, which can have entries in the following formats:
#
#  <key> = <value>
#  <key> {
#     <line1>
#     <line2>
#      ...
#   }
#
# In the latter case the returned value is a single string, which can be split with
# split("\n", ...)
#
##****************************************************************************
    my $file = shift;
    return unless (-r $file);

    my %cfg;
    my $val;
    local $_ = read_file($file);

    my $comment = '#.+';
    s/$comment//mg;		# First clean out any comments 

    check_syntax($_);

    while (/([^\{\}=]+)([=\{])/igsxc) {
	my $key = lc clean_ends($1);
	if ($2 eq '=') {
	    /(.+)/gcxm;
	    $cfg{$key} = clean_ends($1);
	} else {
	    my $depth = 1;
	    my $text = '';
	    while (/ ([^\{\}]*) ([\{\}]) /gcxs) {
		$text .= $1;
		$depth += ($2 eq '{') ? 1 : -1;
		$text .= $2 unless $depth==0;
		last unless $depth;
	    }
	    $cfg{$key} = clean_ends($text);
	}
    }

    # Parse format definition entry into 
    if ($cfg{format_definition}) {
	my $line;
	foreach $line (split "\n", $cfg{format_definition}) {
	    next unless $line =~ /\S/;
	    my ($name, @opt) = split ' ', $line;
	    foreach my $opt (@opt) {
		$cfg{model_format}->{lc $name}->{$1} = $2 if $opt =~ /(\S+)=(\S+)/;
	    }
	}
	delete $cfg{format_definition};
    }

    return %cfg;
}

##****************************************************************************
sub clean_ends {
##****************************************************************************
    my $val = shift;
    $val =~ s/^\s+(\S)/$1/;
    $val =~ s/(\S)\s+$/$1/;
    return $val;
}

##****************************************************************************
sub check_syntax {
##****************************************************************************
#  Very basic syntax check: are there an equal number of { and } ?
    local $_ = shift;
    s/[^{}]//g;
    my $t = length;
    s/[^{]//g;
    die "Unequal number of '{' and '}' in config file\n"
      unless (2 * length == $t);
}
    
# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Ska::Process - Processing utilities for spacecraft/aspect analysis

=head1 SYNOPSIS

 use Ska::Process qw(:all);

    get_archive_files( options.. )
    get_params( $param_file )
    make_local_copy( $in_file, [$out_file], [{copy=>1, absolute=>1}])
    send_mail( options.. )

=head1 METHODS

=over

=item get_archive_files( options.. )

See /proj/sot/ska/share/astromon/get_cat_obs_data.pl or /proj/sot/ska/lib/perl/RunAsp.pm
for examples.

=item get_params( $param_file )

Read IRAF-style parameter file $param_file.

=item make_local_copy( $in_file, [$out_file], [{copy=>1, absolute=>1}])

Make a local copy of or link to the specified in_file.  Gunzip
it if necessary.  If out_file is not supplied, then use the
basename of the in_file.  Don't do anything if the output file is 
already there.  Finally, if the 'copy' option is specified in the
optional hash reference (last argument), then make an actual copy
instead of just linking (if the in_file is not gzipped).

=item send_mail( options.. )

Send an email using Mail::Send as specified by the options:

  addr_list => $address || [$address1, ...],  # scalar or array ref
  subject   => $subject,
  message   => $message,
  loud      => $loud,    # Print mail message to STDOUT (default=0)
  dryrun    => $dryrun;  # Don't send mail (for debugging) (default=0)

Only the Ch2xs -A -X -n Module::Name::Here<addr_list> option is required.

=back

=head1 DEPRECATED METHODS

These package methods are available but are deprecated for any new code.

=over

=item run_tool( $cmd, [{options..}] )

Use Ska::Run instead.

=item message( $message, [options..] )

Use Ska::Message instead.

=item remote_shell_cmd

Figure a better way or write new method.

=item parse_config

Use Config::General instead.

=back

=head1 AUTHOR

Tom Aldcroft

=head1 SEE ALSO

perl(1).

=cut

