# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
sub run_task {
##****************************************************************************
#  Run a processs step task
##****************************************************************************
    my $self = shift;
    my $task = shift;
    my $f;			# Short generic file scalar or array ref var

    if (not $task) {
	message(1, "ERROR - called run_task with no task specification\n");
	return;
    }

    my $task_name = $task->{name} || 'WARNING - process name undefined';

    # Skip task if processing has already stopped, unless 'always_run' is defined.
    # This enables clean-up steps and report generation even for failed processing.
    return if (defined $self->stopped_at and not $task->{always_run});

    # Set values to allow nested looping within task (via build_task_loops())
    my @loop_info;
    while (my ($var, $val) = each %{$task->{loop_val}}) {
	$self->set_var({ var   => $var,
			 value => $val
		       });
	push @loop_info, "$var=$val";
    }

    # Print identifying info for task
    message(3, "\n");
    message(3, "-----------");
    message(2, "  " . ucfirst($task->{type} . 'process') . " step: $task_name ");
    message(2, "for " . join("  ", @loop_info)) if (@loop_info);
    message(2, "\n");
    message(3, "\n");

    if (defined $task->{stop}) {
	message(3, "Processing stopped by 'stop' command\n");
	return ;
    }

    # Change to specified directory within scope of this subroutine
    my $dir;
    if (defined $task->{dir}) {
	my $dir_name = $self->interpolate_string($task->{dir});
	unless (-d $dir_name) {
	    message(3, "WARNING - directory '$dir_name' does not exist from " . cwd() . "\n");
	    return; 
	}
	eval { $dir = io($dir_name) };
	if ($@) {
	    message(1, "ERROR in $task_name - $@\n");
	    return;
	}
	eval { $dir->chdir };
	if ($@) {
	    message(1, "ERROR in $task_name - 'chdir $task->{dir}' failed: $@\n");
	    return;
	}
    }

    # Interpolate the task data structure
    $task = $self->interpolate_struct($task);
    if (not $task) {
	message(1, "ERROR in $task_name - could not interpolate task\n");
	return;
    }

    message(5, Dumper $task);

    # Set environment variables
    foreach my $setenv (listify( $task->{setenv} )) {
	%ENV = GrabEnv::grabenv('csh', "setenv $setenv");
    }

    # Check for "run_if_exists" files.  All these files must be present, but if
    # not then the task exits with success instead of failure.
    # This allows an optional processing step if a certain file or files exist.
    if (defined ($f = $task->{run_if_file})) {
	my $all_files_present = 1;
	for (listify($f)) {
	    my $found = bsd_glob($_, undef);
	    $all_files_present = 0 unless $found;
	    message(4, "Checking for file(s) '$_': " . ($found ? '' : 'not ') . "found\n");
	}
	unless ($all_files_present) {
	    message(3, "Not all 'run_if_file' files found - skipping step\n");
	    return 1 ;
	}
    }
	
    # Check for "run_if_defined" values.  
    # This allows an optional processing step if a certain value exists.
    if (defined ($f = $task->{run_if_defined})) {
	my $all_values_defined = 1;
	for (listify($f)) {
	    my $defined = defined $self->get_var($_);
	    $all_values_defined = 0 unless $defined;
	    message(4, "Checking status of variable '$_': " . ($defined ? '' : 'not ') . "defined\n");
	}
	unless ($all_values_defined) {
	    message(3, "Not all 'run_if_defined' values defined - skipping step\n");
	    return 1 ;
	}
    }
	
    # Check for "run_if_true" values.  
    # This allows an optional processing step if a certain value is true.
    if (defined ($f = $task->{run_if_true})) {
	my $all_values_true = 1;
	for (listify($f)) {
	    $all_values_true = 0 unless $_;
	}
	unless ($all_values_true) {
	    message(3, "Not all 'run_if_true' values true - skipping step\n");
	    return 1 ;
	}
    }
	
    # Check for required input files
    if (defined ($f = $task->{require_file})) {
	my $all_files_present = 1;
	for (listify($f)) {
	    my $found = bsd_glob($_, undef);
	    $all_files_present = 0 unless $found;
	    message(4, "Checking for file(s) '$_': " . ($found ? '' : 'not ') . "found\n");
	}
	unless ($all_files_present) {
	    message(3, "ERROR - Not all 'require_file' files found\n");
	    return;
	}
    }
	
    # Check file dependencies.  Task may not need to be run
    if ((my @target_files = listify($task->{target_file})) > 0) {
	my $target_needs_update;

	# Glob any values in 'depend_glob'.  This allows optional glob dependencies.
	# Any *required* dependencies can be given as a glob in 'depend_file'.
	$task->{depend_file} = [ listify($task->{depend_file}) ];  # force into array_ref
	push @{$task->{depend_file}}, bsd_glob($_, undef) for listify($task->{depend_glob});

	eval { $target_needs_update = test_dep_with_globs($task->{target_file}, $task->{depend_file}) };
	message(5, "Target needs update = $target_needs_update\n");
	if ($@) {
	    message(1, "ERROR in $task_name - test_dep_with_globs died with '$@'");
	    return;
	}
	unless ($target_needs_update) {
	    message(3, "Target does not need update - skipping step\n");
	    return 1;
	}
    }

    # Delete specified files before running command
    if ($f = $task->{delete_file}) {
	foreach (listify($f)) {
	    foreach my $file (bsd_glob($_, undef)) {
		if (-w $file) {
		    message(4, "Deleting file '$_'\n");
		    unlink $file;
		} else {
		    message(4, "WARNING - File '$_' is not writable, cannot delete\n");
		}
	    }
	}
    }

    # Print stuff (mostly for debug)
    foreach (listify($task->{print})) {
	message(3, "$_\n");
    }

    # Actually run the command via shell or Yaxx method
    if (my @command = listify($task->{command})) {
	my $out_all;
	foreach my $command (@command) {
	    $command =~ s/\n/ \\\n/g;
	    my ($error, @out) = Ska::Run::run($command,
					      loud => ($self->verbose_stdout >= 4),
					     );

	    # Send command output to log file but not STDOUT (already done by run())
	    message(4, "Running $command\n", {stdout => 0});
	    message(4, join('',@out), {stdout => 0});

	    if ($error) {
		if ($error =~ /WARNING.+returned non-zero status/) {
		    my $warning = $task->{warn_non_zero} || $error;
		    message(1, "WARNING in $task_name - $warning\n");
		    return;
		}
	    }
	    $out_all .= join('',@out);
	}
	# Allow for "set_var <var>" options
	if (my @set_var = listify($task->{set_var})) {
	    foreach my $var (@set_var) {
		$self->set_var({ var => $var,
				 value => $out_all,
				 chomp => $task->{chomp},
			       });
	    }
	}
    } elsif (my $task_method = $task->{method}) {
	my $ok;
	message(4, "Running method '$task_method'\n");
	eval { $ok = $self->$task_method( $task ) };
	if ($@) {
	    message(1, "ERROR in $task_name - $task_method: $@\n");
	    return;
	}
	unless ($ok) {
	    message(1, "ERROR in $task_name - $task_method returned unsuccessful status\n");
	    return;
	}
    } elsif (my @include = listify($task->{include})) {
	my @include_process_step;
	# Read each include config file and extract any process_step items
	foreach my $include (@include) {
	    message(3, "Reading config include file $include\n");
	    my %cfg = ParseConfig(-ConfigFile => $include, -IncludeRelative => 1);
	    push @include_process_step, listify($cfg{process_step});
	}
	
	# Splice into the process_step list at the right place
	splice(@{$self->{process_step}}, $self->{curr_process_step}+1, 0, @include_process_step);
    }

    # Touch files to help with dependency checks (mostly for files that get updated in place)
    foreach (listify($task->{touch_file})) {
	message(4, "Touching file '$_'\n");
	io($_)->touch;
    }

    # If needed check that the processing actually succeeded in satisfying the dependencies
    if ((my @target_files = listify($task->{target_file})) > 0) {
	my $target_needs_update;
	eval { $target_needs_update = test_dep_with_globs($task->{target_file}, $task->{depend_file}) };
	if ($@) {
	    message(1, "ERROR in $task_name - test_dep_with_globs died with '$@'");
	    return;
	}
	if ($target_needs_update) {
	    message(1, "ERROR in $task_name - processing did not satisfy dependencies\n");
	    return ;
	}
    }

    message(3, "Processing step succeeded\n");
    return 1;
}

##****************************************************************************
sub test_dep_with_globs {
##****************************************************************************
    my $targets =  shift;
    my $depends =  shift;
    my (@targets_out, @depends_out);

    # Glob each target and collect into a master target list.  If any
    # target glob doesn't return at least one file then force action

    my $force = 0;
    foreach my $target (listify($targets)) {
	my @files = bsd_glob($target, undef);
	if (@files) {
	    push @targets_out, @files;
	} else {
	    $force = 1;
	    last;
	}
    }

    # Glob each depend and collect into a master depend list.  If any
    # depend glob doesn't return at least one file then die (which must
    # be caught by eval) to indicate error.  Even if $force is already set
    # still go through and catch depend errors.

    foreach my $depend (listify($depends)) {
	my @files = bsd_glob($depend, undef);
	if (@files) {
	    push @depends_out, @files;
	} else {
	    die "Missing dependency file '$depend'\n";
	}
    }

    # If not already forcing then do a more detailed dependency check (with file dates
    # etc) with test_dep.  Test_dep can also die on error.

    return $force ? 1 : test_dep(-depend => \@depends_out, -target => \@targets_out);
}

##****************************************************************************
sub build_task_loops {
##****************************************************************************
    my $task = shift;

    # If no loops are specified then return a single empty hashref
    return {} unless defined $task->{loop};

    my @loop_vars;
    my %loop_vals;
    foreach my $loop (listify($task->{loop})) {
	my ($var, $vals) = Text::RecordParser->new(data => $loop,
						   field_separator=> qr/\s* = \s*/x,
						   trim=>1,
						  )->fetchrow_array;
	die "Loop specification error - usage: loop <var> = val1 val2 ..\n  Found 'loop $loop\n"
	  unless (defined $var and defined $vals);

	my @vals = Text::RecordParser->new(data => $vals,
					   field_separator=> qr/\s+/x,
					   trim=>1,
					  )->fetchrow_array;
	push @loop_vars, $var unless exists $loop_vals{$var};
	push @{$loop_vals{$var}}, @vals;
    }
    
    # Recursively generate loops corresponding to loop_vars/vals
    my @task_loop;
    set_loop_vals(\@loop_vars, \%loop_vals, {}, \@task_loop, 0);

    return @task_loop;
}

##****************************************************************************
sub set_loop_vals {
##****************************************************************************
    my $vars = shift;
    my $vals = shift;
    my $state = shift;
    my $task_loop = shift;
    my $depth = shift;
    
    if ($depth < @{$vars}) {
	my $var = $vars->[$depth];
	for my $val (@{$vals->{$var}}) {
	    $state->{$var} = $val;
	    set_loop_vals($vars, $vals, $state, $task_loop, $depth+1);
	}
    } else {
	push @{$task_loop}, { %$state };
    }
    return;
}

1;
