# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
# HERE are a handful of accessor methods which ensure that the values are
# properly initialized.  These may throw a fatal error if that is not
# possible.
##****************************************************************************

##****************************************************************************
sub XY {
##****************************************************************************
    my $self = shift;
    my $det = $self->detector;
    if (not defined $det) {
	message(1, "ERROR - attemp to calculate (X,Y) without specifying detector\n");
	confess $Yaxx::FATAL_ERROR;
    }
    
    if (not defined $self->{X}{$det} or not defined $self->{Y}{$det}) {
	if (defined $self->{ra} and defined $self->{dec}) {
	    my $file = q{'} . $self->file('evt') . q{[events]'};
	    my $error = Ska::Run::run("sky2xy infile=$file xsky=$self->{ra} ysky=$self->{dec}",
				      timeout => $self->timeout,
				      loud    => 0,
				     );
	    if ($error) {
		message(1, "ERROR in sky2xy - $error\n");
		confess $Yaxx::FATAL_ERROR;
	    }
	    
	    my @out;
	    ($error, @out) = Ska::Run::run("pget sky2xy xpix ypix",
					   timeout => $self->timeout,
					   loud => 0,
					  );
	    if (not $error and $out[0] =~ /\d/ and $out[1] =~ /\d/) {
		chomp($self->{X}{$det} = $out[0]);
		chomp($self->{Y}{$det} = $out[1]);
	    } else {
		message(1, "ERROR in pget sky2xy - $error\n");
		confess $Yaxx::FATAL_ERROR;
	    }		
	} 

	# After all that, see if we have an X and Y!
	if (not defined $self->{X} or not defined $self->{Y}) {
	    message(1, "ERROR - could not determine source sky (X,Y).  Need\n");
	    message(1, "        (X,Y) or (ra,dec) in source table\n");
	    confess $Yaxx::FATAL_ERROR;
	}
    }

    return ($self->{X}{$det}, $self->{Y}{$det});
}

##****************************************************************************
sub ra_dec {
##****************************************************************************
    my $self = shift;
    my $det = $self->detector;

    if (not defined $self->{ra} or not defined $self->{dec}) {
	if (not defined $det) {
	    message(1, "ERROR - attemp to calculate (ra,dec) from (X,Y) without specifying detector\n");
	    confess $Yaxx::FATAL_ERROR;
	}

	if (defined $self->{X}{$det} and defined $self->{Y}{$det}) {
	    my $file = q{'} . $self->file('evt') . q{[events]'};
	    my $error = Ska::Run::run("xy2sky infile=$file xpix=$self->{X}{$det} ypix=$self->{Y}{$det}",
				      timeout => $self->timeout,
				      loud    => 1,
				     );
	    if ($error) {
		message(1, "ERROR in xy2sky - $error\n");
		confess $Yaxx::FATAL_ERROR;
	    }
	    
	    my @out;
	    ($error, @out) = Ska::Run::run("pget xy2sky xsky ysky",
					   timeout => $self->timeout,
					   loud => 1,
					  );
	    if (not $error and $out[0] =~ /\d/ and $out[1] =~ /\d/) {
		chomp($self->{ra}  = $out[0]);
		chomp($self->{dec} = $out[1]);
	    } else {
		message(1, "ERROR in pget xy2sky - $error\n");
		confess $Yaxx::FATAL_ERROR;
	    }		
	} 

	# After all that, see if we have an X and Y!
	if (not defined $self->{ra} or not defined $self->{dec}) {
	    message(1, "ERROR - could not determine source RA,Dec\n");
	    confess $Yaxx::FATAL_ERROR;
	}
    }

    return ($self->{ra}, $self->{dec});
}

##****************************************************************************
sub rad {
# Source extraction radius
##****************************************************************************
    my $self = shift;
    if (@_) {
	$self->{rad} = $_[0];
    }

    if (not defined $self->{rad}) {
	# No radius spec'd, so determine using the Chandra circular ECF values
	# determined by the cal group.  If an ECF object hasn't yet been
	# created, do so now

	$self->{rad} = 800;

#	my $msg = sprintf("Extraction radius (%.0f%%) is %.2f pixels for (theta,phi)=(%.2f',%.1f deg)\n",
#			  $self->psf_fraction * 100, $self->{rad}, $out->{theta}, $out->{phi});
#	message($msg);

	if ($self->{rad} < $self->{min_src_rad}) {
	    $self->{rad} = $self->{min_src_rad};
	    message(3, "Expanding extraction radius to $self->{min_src_rad} pixels\n");
	}
    }

    return $self->{rad};
}


1;
